/*
 * server.h: header for server.c 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 *
 * @(#)$Id: server.h,v 1.7 1994/07/02 02:38:10 mrg Exp $
 */

#ifndef _SERVER_H_
#define _SERVER_H_
  
#include "irc_std.h"
#include "whois.h" /* for WhoisQueue */
#include "who.h" /* For who queue */

struct notify_stru;

/*
 * type definition to distinguish different
 * server versions
 */
#define Server2_8	1
#define Server2_9	2
#define Server2_10	3
#define Server_u2_8	4
#define Server_u2_9	5
#define Server_u2_10	6
#define Server_u3_0	7

/* Server: a structure for the server_list */
typedef	struct
{
	char	*name;			/* the name of the server */
	char	*itsname;		/* the server's idea of its name */
	char	*password;		/* password for that server */
	int	port;			/* port number on that server */
	char	*nickname;		/* nickname for this server */
	char	*away;			/* away message for this server */
	int	operator;		/* true if operator */
	int	version;		/* the version of the server -
					 * defined above */
	int	server2_8;		/* defined if we get an 001 numeric */
	char	*version_string;	/* what is says */
	int	whois;			/* true if server sends numeric 318 */
	int	flags;			/* Various flags */
	char	umode[27];		/* User mode */
	int	s_takes_arg;		/* Set to 1 if s user mode has arg */
	int	connected;		/* true if connection is assured */
	int	write;			/* write descriptor */
	int	read;			/* read descriptior */
	pid_t	pid;			/* process id of server */
	int	eof;			/* eof flag for server */
	int	motd;			/* motd flag (used in notice.c) */
	int	sent;			/* set if something has been sent,
					 * used for redirect */
	char	*buffer;		/* buffer of what dgets() doesn't get */
	WhoisQueue	*WQ_head;	/* WHOIS Queue head */
	WhoisQueue	*WQ_tail;	/* WHOIS Queue tail */
	WhoisStuff	whois_stuff;	/* Whois Queue current collection buffer */
	WhoEntry *	who_queue;	/* Who queue */
	IsonEntry *	ison_queue;	/* Ison queue */

	struct in_addr local_addr;      /* ip address of this connection */
	struct notify_stru *notify_list;	/* Notify list for this server */
}	Server;

extern	int	attempting_to_connect;
extern	int	number_of_servers;
extern	int	connected_to_server;
extern	int	never_connected;
extern	int	using_server_process;
extern	int	primary_server;
extern	int	from_server;
extern	int	last_server;
extern	char	*connect_next_nick;
extern	int	parsing_server_index;

typedef	unsigned	short	ServerType;

extern	void	add_to_server_list 	(char *, int, char *, char *, int);
extern	void	build_server_list 		(char *);
extern	int	connect_to_server 		(char *, int, int);
extern	int	connect_to_server_by_refnum 	(int, int);
extern	void	get_connected 			(int);
extern	int	read_server_file 		(void);
extern	void	display_server_list 		(void);
extern	int	server_list_size 		(void);
extern	void	send_to_server 			(const char *, ...) __A(1);
extern	WhoisStuff	*get_server_whois_stuff (int);
extern	WhoisQueue	*get_server_qhead 	(int);
extern	WhoisQueue	*get_server_qtail 	(int);
extern	void	servercmd			(char *, char *, char *);
extern	char	*get_server_nickname 		(int);
extern	char	*get_server_name 		(int);
extern	char	*get_server_itsname 		(int);
extern	void	set_server_flag 		(int, int, int);
extern	int	find_in_server_list		(char *, int);
extern	char	*create_server_list 		(void);
extern	void	set_server_motd 		(int, int);
extern	int	get_server_motd 		(int);
extern	int	get_server_operator 		(int);
extern	int	get_server_version 		(int);
extern	void	close_server 			(int, char *);
extern	void	MarkAllAway 			(char *, char *, char *);
extern	int	is_server_connected 		(int);
extern	void	flush_server 			(void);
extern	void	set_server2_8 			(int, int);
extern	int	get_server2_8 			(int);
extern	int	get_server_flag 		(int, int);
extern	void	set_server_operator 		(int, int);
extern	void	server_is_connected 		(int, int);
extern	int	parse_server_index 		(char *);
extern	void	parse_server_info 	(char *, char **, char **, char **);
extern	void	set_server_bits 	(fd_set *);
extern	void	set_server_itsname 	(int, char *);
extern	void	set_server_version 	(int, int);
extern	int	is_server_open 		(int);
extern	int	get_server_port 	(int);
extern	char	*set_server_password 	(int, char *);
extern	void	set_server_nickname 	(int, char *);
extern	void	set_server_qhead 	(int, WhoisQueue *);
extern	void	set_server_qtail 	(int, WhoisQueue *);
extern	void	close_all_server 	(void);
extern	void	disconnectcmd 		(char *, char *, char *);
extern	char 	*get_umode 		(int);
extern	int	find_server_refnum 	(char *, char **);

/* server_list: the list of servers that the user can connect to,etc */
extern	Server	*server_list;
extern	const char	*umodes;

#endif /* _SERVER_H_ */
