/*
 * whois.c: Some tricky routines for querying the server for information
 * about a nickname using WHOIS.... all the time hiding this from the user.  
 *
 * Probably 99% of this file is obsolete now that we dont use 'whois' to
 * get userhost information (this method was required way back in 2.5/2.6
 * servers which didnt provide userhost info on each incoming line).  The
 * only thing that is still used is the USERHOST and ISON commands which 
 * are used for well, the /userhost and /ison commands.  Those things can
 * probably be handled in a better method then this hackery, and i may do
 * that.
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: whois.c,v 1.14 1994/07/02 02:32:13 mrg Exp $";
#endif

#include "irc.h"
#include "whois.h"
#include "hook.h"
#include "lastlog.h"
#include "vars.h"
#include "server.h"
#include "ignore.h"
#include "ircaux.h"
#include "notify.h"
#include "numbers.h"
#include "window.h"
#include "commands.h"
#include "output.h"
#include "parse.h"
#include "ctcp.h"

#include <stdarg.h>

char	whois_nick[] = "#WHOIS#";
char	wait_nick[] = "#WAIT#";
char	redirect_nick[] = "#RED#";

/* current setting for BEEP_ON_MSG */
int	beep_on_level;

static	int	ignore_whois_crap = 0;
static	int	eat_away = 0;


/* WQ_head and WQ_tail point to the head and tail of the whois queue */
WhoisQueue *WQ_head = (WhoisQueue *) 0;
WhoisQueue *WQ_tail = (WhoisQueue *) 0;


/* This does NOT belong here */
void	set_beep_on_msg (char *str)
{
	beep_on_level = parse_lastlog_level(str);
	set_string_var(BEEP_ON_MSG_VAR, bits_to_lastlog_level(beep_on_level));
}

/*
 * whois_queue_head: returns the nickname at the head of the whois queue, or
 * NULL if the queue is empty.  It does not modify the queue in any way. 
 */
static char	*whois_queue_head (void)
{
	if ((WQ_head = (WhoisQueue *) get_server_qhead(from_server)) != NULL)
		return (WQ_head->nick);
	else
		return ((char *) 0);
}

static void	(*whois_func_head()) (void)
{
	if ((WQ_head=(WhoisQueue *) get_server_qhead(from_server)) != NULL)
		return (WQ_head->func);
	else
		return NULL;
}

/*
 * remove_from_whois_queue: removes the top element of the whois queue and
 * returns the element as its function value.  This routine repairs handles
 * all queue stuff, but the returned element is mallocd and must be freed by
 * the calling routine 
 */
static	WhoisQueue *remove_from_whois_queue (void)
{
	WhoisQueue *new;

	new = (WhoisQueue *) get_server_qhead(from_server);
	set_server_qhead(from_server, new->next);
	if (new->next == (WhoisQueue *) 0)
		set_server_qtail(from_server, (WhoisQueue *) 0);
	return (new);
}

/*
 * clean_whois_queue: this empties out the whois queue.  This is used after
 * server reconnection to assure that no bogus entries are left in the whois
 * queue 
 */
void	clean_whois_queue (void)
{
	WhoisQueue *thing;

	while (whois_queue_head())
	{
		thing = remove_from_whois_queue();
		new_free(&thing->nick);
		new_free(&thing->text);
		new_free((char **)&thing);
	}
	ignore_whois_crap = 0;
	eat_away = 0;
}

void	whois_userhost (char *nick, void (*func)(), char *cmd)
{
	WhoisQueue *new;

	if ((!nick || !*nick) || ! is_server_connected(from_server))
		return;

	new = (WhoisQueue *) new_malloc(sizeof(WhoisQueue));
	new->nick = m_strdup(nick);
	new->text = m_strdup(cmd);
	new->func = func;
	new->next = (WhoisQueue *) 0;

	if ((void *) get_server_qhead(from_server) == (void *) 0)
		set_server_qhead(from_server, new);
	if (get_server_qtail(from_server))
		((WhoisQueue *) get_server_qtail(from_server))->next = new;
	set_server_qtail(from_server, new);
	send_to_server("USERHOST %s", nick);
}



/* 
 * userhost_returned: this is called when numeric 302 is received in
 * numbers.c. USERHOST must also remain the property of the WHOIS
 * queue. Sending it without going via the WHOIS queue will cause
 * the queue to become corrupted.
 *
 * While USERHOST gives us similar information to WHOIS, this routine's
 * format is a little different to those that handle the WHOIS numerics.
 * A list of nicks can be supplied to USERHOST, and any of those
 * nicks are not currently signed on, it will just omit reporting it.
 * this means we must go through the list of nicks returned and check
 * them for missing entries. This means stepping through the requested
 * nicks one at a time.
 *
 * A side effect of this is that user initiated USERHOST requests get
 * reported as separate replies even though one request gets made and
 * only one reply is actually received. This should make it easier for
 * people wanting to use USERHOST for their own purposes.
 *
 * In this routine, cnick points to all the information in the next
 * entry from the returned data.
 */
void	userhost_returned (char *from, char **ArgList)
{
	WhoisQueue *thing;
	WhoisStuff *whois_stuff = NULL;
	char	*nick,
		*cnick = NULL,
		*tnick;
	char	*user,
		*cuser = (char *) 0;
	char	*host,
		*chost = (char *) 0;
	int	ishere,
		cishere = 1;
	int	isoper,
		cisoper = 0;
	int	noton,
		isuser,
		parsed;
	char	*queue_nicks;

	if (!ArgList[0])
		return;

	isuser = (whois_func_head() == USERHOST_USERHOST);
	whois_stuff = get_server_whois_stuff(from_server);
	thing = remove_from_whois_queue();
	queue_nicks = thing->nick;

	parsed = 0;
	while ((parsed || (cnick = next_arg(ArgList[0], ArgList))) || queue_nicks)
	{
		if (queue_nicks)
		{
			tnick = next_arg(queue_nicks, &queue_nicks);
			if (!*queue_nicks)
				queue_nicks = (char *) 0;
		}
		else
			tnick = NULL;
		if (cnick && !parsed)
		{
			if (!(cuser = index(cnick,'=')))
				break;
			if (*(cuser - 1) == '*')
			{
				*(cuser - 1) = '\0';
				cisoper = 1;
			}
			else
				cisoper = 0;
			*cuser++ = '\0';
			if (*cuser++ == '+')
				cishere = 1;
			else
				cishere = 0;
			if (!(chost = index(cuser, '@')))
				break;
			*chost++ = '\0';
			parsed = 1;
		}
		if (!cnick || (tnick && my_stricmp(cnick, tnick)))
		{
			if (tnick)
				nick = tnick;
			else
				nick = cnick;
			user = host = "<UNKNOWN>";
			isoper = 0;
			ishere = 1;
			noton = 1;
		}
		else
		{
			nick = cnick;
			user = cuser;
			host = chost;
			isoper = cisoper;
			ishere = cishere;
			noton = parsed = 0;
		}
		if (!isuser)
		{
			malloc_strcpy(&whois_stuff->nick, nick);
			malloc_strcpy(&whois_stuff->user, user);
			malloc_strcpy(&whois_stuff->host, host);
			whois_stuff->oper = isoper;
			whois_stuff->not_on = noton;
			if (!ishere)
				malloc_strcpy(&whois_stuff->away, empty_string);
			else
				new_free(&whois_stuff->away);
			thing->func(whois_stuff, tnick, thing->text);
			new_free(&whois_stuff->away);
		}
		else
		{
			if (do_hook(current_numeric, "%s %s %s %s %s", nick,
			    isoper ? "+" : "-", ishere ? "+" : "-", user, host))
				put_it("%s %s is %s@%s%s%s", numeric_banner(),
					nick, user, host, isoper ?
					" (Is an IRC operator)" : empty_string,
					ishere ? empty_string : " (away)");
		}
	}
	if (thing)
	{
		new_free(&thing->nick);
		new_free(&thing->text);
		new_free((char **)&thing);
	}
}

void	userhost_cmd_returned (WhoisStuff *stuff, char *nick, char *text)
{
	char	args[BIG_BUFFER_SIZE + 1];

	strcpy(args, stuff->nick ? stuff->nick : empty_string);
	strcat(args, stuff->oper ? " + " : " - ");
	strcat(args, stuff->away ? "+ " : "- ");
	strcat(args, stuff->user ? stuff->user : empty_string);
	strcat(args, space);
	strcat(args, stuff->host ? stuff->host : empty_string);
	parse_line("USERHOST", text, args, 0, 0);
}

